#include <LiquidCrystal_I2C.h>

// Input signals
#define SHOOT_BUTTON 12
#define RESET_BUTTON 13
#define IR_RECEIVER A0

// Output signals
#define IR_LED 5
#define BUZZER 9

// LCD variables
LiquidCrystal_I2C lcd(0x27, 16, 2);	// I2C address 0x27, 16 column and 2 rows
unsigned long previousMillisLCD = 0;	// Store last time LCD was updated
const long intervalLCD = 500;		// Interval at which to update LCD (milliseconds)

// Gun variables
int lives = 3;
int ammo = 10;
unsigned long previousMillisShot = 0;	// Store last time gun was shot
const long intervalShot = 1000;		// Minimum interval between shots

// IR Sensor variables
int ambientRadiation = 0;		// Variable ambiental infrared radiation
int lightLevel = 0;			// IR radiation when button is pressed
const int lightThreshhold = 4;		// Threshhold used to determine whether the gun is pointed at the target

// Buzzer music, credit to https://github.com/hibit-dev/buzzer/blob/master/src/movies/game_of_thrones/game_of_thrones.ino
#define NOTE_C4  262
#define NOTE_DS4 311
#define NOTE_F4  349
#define NOTE_G4  392
#define REST      0

int tempo = 85;				// Make the song slower or faster

int melody[] = {
  NOTE_G4,8, NOTE_C4,8, NOTE_DS4,16, NOTE_F4,16  
};

int notes = sizeof(melody) / sizeof(melody[0]) / 2;

// Duration of a whole note in ms
int wholenote = (60000 * 4) / tempo;
int divider = 0, noteDuration = 0;

void playTone() {
  // Iterate over the notes of the melody
  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {
    // Calculate the duration of each note
    divider = melody[thisNote + 1];
    noteDuration = (wholenote) / divider;

    // Play the note for 90% of the duration, leaving 10% as a pause
    tone(BUZZER, melody[thisNote], noteDuration * 0.9);

    // Wait for the specied duration before playing the next note
    delay(noteDuration);

    // stop the waveform generation before the next note.
    noTone(BUZZER);
  }
}

void showShotMessage() {
  lcd.setCursor(0, 0);
  lcd.print("##  You got   ##");
  lcd.setCursor(0, 1);
  lcd.print("##   shot!    ##");
}

void showDeadMessage() {
  unsigned long currentMillis = millis();

  if (currentMillis - previousMillisLCD >= intervalLCD) {
    // save the last time the LCD was updated
    previousMillisLCD = currentMillis;

    lcd.setCursor(0, 0);
    lcd.print("##  You are   ##");
    lcd.setCursor(0, 1);
    lcd.print("##   dead...  ##");
  }
}

void showChargeMessage() {
  unsigned long currentMillis = millis();

  if (currentMillis - previousMillisLCD >= intervalLCD) {
    // save the last time the LCD was updated
    previousMillisLCD = currentMillis;

    lcd.setCursor(0, 0);
    lcd.print("##   Please   ##");
    lcd.setCursor(0, 1);
    lcd.print("##  recharge  ##");
  }
}

void showGunStatus() {
  unsigned long currentMillis = millis();

  if (currentMillis - previousMillisLCD >= intervalLCD) {
    // save the last time the LCD was updated
    previousMillisLCD = currentMillis;

    lcd.setCursor(0, 0);
    lcd.print("##  Ammo      ##");
    lcd.setCursor(11, 0);
    lcd.print(ammo);
    lcd.setCursor(0, 1);
    lcd.print("##  Lives     ##");
    lcd.setCursor(11, 1);
    lcd.print(lives);
  }
}

void serialDebugMessages() {
  // Debug messages
  Serial.print(" *** AMMO ");
  Serial.print(ammo);
  Serial.print(" *** LIVES ");
  Serial.print(lives);
  Serial.print(" *** lightLevel ");
  Serial.print(lightLevel);
  Serial.print(" *** ambientRadiation ");
  Serial.print(ambientRadiation);
  Serial.print("\n");
}

void setup() {
  pinMode(SHOOT_BUTTON, INPUT_PULLUP);
  pinMode(RESET_BUTTON, INPUT_PULLUP);
  pinMode(IR_RECEIVER, INPUT);

  pinMode(IR_LED, OUTPUT);
  pinMode(BUZZER, OUTPUT);

  // Initialize the LCD
  lcd.begin(16, 2);
  lcd.init();
  lcd.backlight();

  // Use serial monitor for debugging
  Serial.begin(9600);

  // Initialize ADC
  ADMUX = 0;
  // ADC1 - channel 1
  ADMUX |= (1 << MUX0);
  // AVCC with external capacitor at AREF pin
  ADMUX |= (1 << REFS0);

  ADCSRA = 0;
  // Set prescaler at 128
  ADCSRA |= (7 << ADPS0);
  // Enable ADC
  ADCSRA |= (1 << ADEN);
}

uint16_t getAnalogData() {
  // Start conversion
  ADCSRA |= (1 << ADSC);

  // Wait until conversion is complete
  while (!(ADCSRA & (1 << ADIF)));

  uint16_t result = ADC;
  return result;
}

void loop() {
  if (ammo > 0 && lives > 0) {
    showGunStatus();

    // Get ambient signal from the IR receiver
    ambientRadiation = getAnalogData();
    
    // Read signal from the button and shoot
    if (digitalRead(SHOOT_BUTTON) == LOW) {
      unsigned long currentMillis = millis();

      if (currentMillis - previousMillisShot >= intervalShot) {
        // save the last time the LCD was updated
        previousMillisShot = currentMillis;

        ammo = ammo - 1;
        digitalWrite(IR_LED, HIGH);
        showGunStatus();
        
        // Read signal from the IR receiver
        lightLevel = analogRead(IR_RECEIVER);
      }

    } else {
      digitalWrite(IR_LED, LOW);
    }
    
    // serialDebugMessages();

    if(lightLevel - ambientRadiation >= lightThreshhold) {
      // The gun was shot
      showShotMessage();
      playTone();
      lives = lives - 1;
      showGunStatus();
      lightLevel = 0;
    }
   
  } else if (ammo == 0) {
    // The player ran out of ammo
    digitalWrite(IR_LED, LOW);
    showChargeMessage();
    
  } else if (lives == 0) {
    // The player lost
    digitalWrite(IR_LED, LOW);
    showDeadMessage();
  }
  
  // Read signal from the button that resets the gun
  if (digitalRead(RESET_BUTTON) == LOW) {
    ammo = 10;
    lives = 3;
  }
}